from veros import distributed, veros_routine, veros_kernel, runtime_state as rst
from veros.variables import allocate

from veros.core import utilities
from veros.core.operators import update, update_add, at, numpy as npx
from veros.core.streamfunction.solvers.base import LinearSolver


@veros_kernel(static_args=("solve_fun",))
def solve_kernel(state, rhs, x0, boundary_val, solve_fun):
    vs = state.variables

    rhs_global = distributed.gather(rhs, state.dimensions, ("xt", "yt"))
    x0_global = distributed.gather(x0, state.dimensions, ("xt", "yt"))

    boundary_mask = ~npx.any(vs.boundary_mask, axis=2)
    boundary_mask_global = distributed.gather(boundary_mask, state.dimensions, ("xt", "yt"))

    if boundary_val is None:
        boundary_val = x0_global
    else:
        boundary_val = distributed.gather(boundary_val, state.dimensions, ("xt", "yt"))

    if rst.proc_rank == 0:
        linear_solution = solve_fun(rhs_global, x0_global, boundary_mask_global, boundary_val)
    else:
        linear_solution = npx.empty_like(rhs)

    return distributed.scatter(linear_solution, state.dimensions, ("xt", "yt"))


class JAXSciPySolver(LinearSolver):
    @veros_routine(
        local_variables=("hvr", "hur", "dxu", "dxt", "dyu", "dyt", "cosu", "cost", "boundary_mask"),
        dist_safe=False,
    )
    def __init__(self, state):
        from jax.scipy.sparse.linalg import bicgstab

        settings = state.settings

        matrix_diags, offsets = self._assemble_poisson_matrix(state)
        jacobi_precon = self._jacobi_preconditioner(state, matrix_diags)
        matrix_diags = tuple(jacobi_precon * diag for diag in matrix_diags)

        @veros_kernel
        def linear_solve(rhs, x0, boundary_mask, boundary_val):
            x0 = utilities.enforce_boundaries(x0, settings.enable_cyclic_x, local=True)
            rhs = npx.where(boundary_mask, rhs, boundary_val)  # set right hand side on boundaries

            def matmul(rhs):
                nx, ny = rhs.shape

                res = npx.zeros_like(rhs)
                for diag, (di, dj) in zip(matrix_diags, offsets):
                    assert diag.shape == (nx, ny)
                    i_s = min(max(di, 0), nx - 1)
                    i_e = min(max(nx + di, 1), nx)
                    j_s = min(max(dj, 0), ny - 1)
                    j_e = min(max(ny + dj, 1), ny)

                    i_s_inv = nx - i_e
                    i_e_inv = nx - i_s
                    j_s_inv = ny - j_e
                    j_e_inv = ny - j_s

                    res = update_add(
                        res,
                        at[i_s_inv:i_e_inv, j_s_inv:j_e_inv],
                        diag[i_s_inv:i_e_inv, j_s_inv:j_e_inv] * rhs[i_s:i_e, j_s:j_e],
                    )

                return res

            linear_solution, _ = bicgstab(
                matmul,
                rhs[2:-2, 2:-2] * self._rhs_scale,
                x0=x0[2:-2, 2:-2],
                tol=0,
                atol=1e-8,
                maxiter=10_000,
            )

            return update(rhs, at[2:-2, 2:-2], linear_solution)

        self._linear_solve = linear_solve
        self._rhs_scale = jacobi_precon

    def solve(self, state, rhs, x0, boundary_val=None):
        """
        Main solver for streamfunction. Solves a 2D Poisson equation. Uses jax.scipy.sparse.linalg
        linear solvers.

        Arguments:
            rhs: Right-hand side vector
            x0: Initial guess
            boundary_val: Array containing values to set on boundary elements. Defaults to `x0`.

        """
        if rst.proc_rank == 0:
            linear_solve = self._linear_solve
        else:
            linear_solve = None

        return solve_kernel(state, rhs, x0, boundary_val, linear_solve)

    @staticmethod
    def _jacobi_preconditioner(state, matrix_diags):
        """
        Construct a simple Jacobi preconditioner
        """
        eps = 1e-20
        main_diag = matrix_diags[0]
        precon = npx.where(npx.abs(main_diag) > eps, 1.0 / main_diag, 1.0)
        return precon

    @staticmethod
    def _assemble_poisson_matrix(state):
        """
        Construct a sparse matrix based on the stencil for the 2D Poisson equation.
        """
        vs = state.variables
        settings = state.settings

        boundary_mask = ~npx.any(vs.boundary_mask[2:-2, 2:-2], axis=2)

        # assemble diagonals
        main_diag = allocate(state.dimensions, ("xu", "yu"), fill=1, local=False, include_ghosts=False)
        east_diag, west_diag, north_diag, south_diag = (
            allocate(state.dimensions, ("xu", "yu"), local=False, include_ghosts=False) for _ in range(4)
        )
        main_diag = (
            -vs.hvr[3:-1, 2:-2]
            / vs.dxu[2:-2, npx.newaxis]
            / vs.dxt[3:-1, npx.newaxis]
            / vs.cosu[npx.newaxis, 2:-2] ** 2
            - vs.hvr[2:-2, 2:-2]
            / vs.dxu[2:-2, npx.newaxis]
            / vs.dxt[2:-2, npx.newaxis]
            / vs.cosu[npx.newaxis, 2:-2] ** 2
            - vs.hur[2:-2, 2:-2]
            / vs.dyu[npx.newaxis, 2:-2]
            / vs.dyt[npx.newaxis, 2:-2]
            * vs.cost[npx.newaxis, 2:-2]
            / vs.cosu[npx.newaxis, 2:-2]
            - vs.hur[2:-2, 3:-1]
            / vs.dyu[npx.newaxis, 2:-2]
            / vs.dyt[npx.newaxis, 3:-1]
            * vs.cost[npx.newaxis, 3:-1]
            / vs.cosu[npx.newaxis, 2:-2]
        )
        east_diag = (
            vs.hvr[3:-1, 2:-2] / vs.dxu[2:-2, npx.newaxis] / vs.dxt[3:-1, npx.newaxis] / vs.cosu[npx.newaxis, 2:-2] ** 2
        )
        west_diag = (
            vs.hvr[2:-2, 2:-2] / vs.dxu[2:-2, npx.newaxis] / vs.dxt[2:-2, npx.newaxis] / vs.cosu[npx.newaxis, 2:-2] ** 2
        )
        north_diag = (
            vs.hur[2:-2, 3:-1]
            / vs.dyu[npx.newaxis, 2:-2]
            / vs.dyt[npx.newaxis, 3:-1]
            * vs.cost[npx.newaxis, 3:-1]
            / vs.cosu[npx.newaxis, 2:-2]
        )
        south_diag = (
            vs.hur[2:-2, 2:-2]
            / vs.dyu[npx.newaxis, 2:-2]
            / vs.dyt[npx.newaxis, 2:-2]
            * vs.cost[npx.newaxis, 2:-2]
            / vs.cosu[npx.newaxis, 2:-2]
        )

        if settings.enable_cyclic_x:
            # couple edges of the domain
            wrap_diag_east, wrap_diag_west = (
                allocate(state.dimensions, ("xu", "yu"), local=False, include_ghosts=False) for _ in range(2)
            )
            wrap_diag_east = update(wrap_diag_east, at[0, :], west_diag[0, :] * boundary_mask[0, :])
            wrap_diag_west = update(wrap_diag_west, at[-1, :], east_diag[-1, :] * boundary_mask[-1, :])
            west_diag = update(west_diag, at[0, :], 0.0)
            east_diag = update(east_diag, at[-1, :], 0.0)

        main_diag = main_diag * boundary_mask
        main_diag = npx.where(main_diag == 0.0, 1.0, main_diag)

        # construct sparse matrix diagonals
        cf = (
            main_diag,
            boundary_mask * east_diag,
            boundary_mask * west_diag,
            boundary_mask * north_diag,
            boundary_mask * south_diag,
        )
        offsets = ((0, 0), (1, 0), (-1, 0), (0, 1), (0, -1))
        if settings.enable_cyclic_x:
            cf += (wrap_diag_east, wrap_diag_west)
            offsets += ((settings.nx, 0), (-settings.nx, 0))

        return cf, offsets



import os

from petsc4py import PETSc
import numpy as onp

from veros import logger, veros_kernel, runtime_settings as rs, runtime_state as rst
from veros.core import utilities
from veros.core.streamfunction.solvers.base import LinearSolver
from veros.core.operators import numpy as npx, update, update_add, at, flush


class PETScSolver(LinearSolver):
    def __init__(self, state):
        if rst.proc_num > 1 and rs.device == "cpu" and "OMP_NUM_THREADS" not in os.environ:
            logger.warning(
                "Environment variable OMP_NUM_THREADS is not set, which can lead to severely "
                "degraded performance when MPI is used."
            )

        settings = state.settings

        if settings.enable_cyclic_x:
            boundary_type = ("periodic", "ghosted")
        else:
            boundary_type = ("ghosted", "ghosted")

        self._da = PETSc.DMDA().create(
            [settings.nx, settings.ny],
            stencil_width=1,
            stencil_type="star",
            comm=rs.mpi_comm,
            proc_sizes=rs.num_proc,
            boundary_type=boundary_type,
            ownership_ranges=[
                (settings.nx // rs.num_proc[0],) * rs.num_proc[0],
                (settings.ny // rs.num_proc[1],) * rs.num_proc[1],
            ],
        )

        if rs.device == "gpu":
            self._da.setVecType("cuda")
            self._da.setMatType("aijcusparse")

        self._matrix, self._boundary_fac = self._assemble_poisson_matrix(state)

        petsc_options = PETSc.Options()

        # setup krylov method
        self._ksp = PETSc.KSP()
        self._ksp.create(self._da.comm)
        self._ksp.setOperators(self._matrix)

        self._ksp.setType("bcgs")
        self._ksp.setTolerances(atol=1e-24, rtol=1e-14, max_it=1000)

        # preconditioner
        self._ksp.getPC().setType("gamg")
        petsc_options["pc_gamg_type"] = "agg"
        petsc_options["pc_gamg_reuse_interpolation"] = True
        petsc_options["pc_gamg_threshold"] = 1e-4
        petsc_options["pc_gamg_sym_graph"] = True
        petsc_options["pc_gamg_agg_nsmooths"] = 2
        petsc_options["mg_levels_pc_type"] = "jacobi"

        if rs.petsc_options:
            petsc_options.insertString(rs.petsc_options)

        self._ksp.setFromOptions()
        self._ksp.getPC().setFromOptions()

        self._rhs_petsc = self._da.createGlobalVec()
        self._sol_petsc = self._da.createGlobalVec()

    def _petsc_solver(self, rhs, x0):
        # hangs on multi-GPU without this
        flush()

        self._da.getVecArray(self._rhs_petsc)[...] = rhs[2:-2, 2:-2]
        self._da.getVecArray(self._sol_petsc)[...] = x0[2:-2, 2:-2]

        self._ksp.solve(self._rhs_petsc, self._sol_petsc)

        info = self._ksp.getConvergedReason()
        iterations = self._ksp.getIterationNumber()

        if info < 0:
            logger.warning(f"Streamfunction solver did not converge after {iterations} iterations (error code: {info})")

        if rs.monitor_streamfunction_residual:
            # re-use rhs vector to store residual
            rhs_norm = self._rhs_petsc.norm(PETSc.NormType.NORM_2)
            self._matrix.multAdd(self._sol_petsc, -self._rhs_petsc, self._rhs_petsc)
            residual_norm = self._rhs_petsc.norm(PETSc.NormType.NORM_2)
            rel_residual = residual_norm / rhs_norm

            if rel_residual > 1e-8:
                logger.warning(
                    f"Streamfunction solver did not achieve required precision (rel. residual: {rel_residual:.2e})"
                )

        return npx.asarray(self._da.getVecArray(self._sol_petsc)[...])

    def solve(self, state, rhs, x0, boundary_val=None):
        """
        Arguments:
            rhs: Right-hand side vector
            x0: Initial guess
            boundary_val: Array containing values to set on boundary elements. Defaults to `x0`.
        """
        rhs, x0 = prepare_solver_inputs(state, rhs, x0, boundary_val, self._boundary_fac)
        linear_solution = self._petsc_solver(rhs, x0)
        return update(rhs, at[2:-2, 2:-2], linear_solution)

    def _assemble_poisson_matrix(self, state):
        """
        Construct a sparse matrix based on the stencil for the 2D Poisson equation.
        """
        vs = state.variables
        settings = state.settings

        matrix = self._da.getMatrix()

        boundary_mask = ~npx.any(vs.boundary_mask[2:-2, 2:-2], axis=2)

        # assemble diagonals
        main_diag = (
            -vs.hvr[3:-1, 2:-2]
            / vs.dxu[2:-2, npx.newaxis]
            / vs.dxt[3:-1, npx.newaxis]
            / vs.cosu[npx.newaxis, 2:-2] ** 2
            - vs.hvr[2:-2, 2:-2]
            / vs.dxu[2:-2, npx.newaxis]
            / vs.dxt[2:-2, npx.newaxis]
            / vs.cosu[npx.newaxis, 2:-2] ** 2
            - vs.hur[2:-2, 2:-2]
            / vs.dyu[npx.newaxis, 2:-2]
            / vs.dyt[npx.newaxis, 2:-2]
            * vs.cost[npx.newaxis, 2:-2]
            / vs.cosu[npx.newaxis, 2:-2]
            - vs.hur[2:-2, 3:-1]
            / vs.dyu[npx.newaxis, 2:-2]
            / vs.dyt[npx.newaxis, 3:-1]
            * vs.cost[npx.newaxis, 3:-1]
            / vs.cosu[npx.newaxis, 2:-2]
        )
        east_diag = (
            vs.hvr[3:-1, 2:-2] / vs.dxu[2:-2, npx.newaxis] / vs.dxt[3:-1, npx.newaxis] / vs.cosu[npx.newaxis, 2:-2] ** 2
        )
        west_diag = (
            vs.hvr[2:-2, 2:-2] / vs.dxu[2:-2, npx.newaxis] / vs.dxt[2:-2, npx.newaxis] / vs.cosu[npx.newaxis, 2:-2] ** 2
        )
        north_diag = (
            vs.hur[2:-2, 3:-1]
            / vs.dyu[npx.newaxis, 2:-2]
            / vs.dyt[npx.newaxis, 3:-1]
            * vs.cost[npx.newaxis, 3:-1]
            / vs.cosu[npx.newaxis, 2:-2]
        )
        south_diag = (
            vs.hur[2:-2, 2:-2]
            / vs.dyu[npx.newaxis, 2:-2]
            / vs.dyt[npx.newaxis, 2:-2]
            * vs.cost[npx.newaxis, 2:-2]
            / vs.cosu[npx.newaxis, 2:-2]
        )

        main_diag = npx.where(boundary_mask, main_diag, 0.0)
        main_diag = npx.where(main_diag == 0.0, 1.0, main_diag)

        # construct sparse matrix
        cf = tuple(
            # copy to NumPy for speed
            onp.asarray(diag)
            for diag in (
                main_diag,
                boundary_mask * east_diag,
                boundary_mask * west_diag,
                boundary_mask * north_diag,
                boundary_mask * south_diag,
            )
        )

        row = PETSc.Mat.Stencil()
        col = PETSc.Mat.Stencil()

        ij_offsets = [(0, 0), (1, 0), (-1, 0), (0, 1), (0, -1)]

        (i0, i1), (j0, j1) = self._da.getRanges()
        for j in range(j0, j1):
            for i in range(i0, i1):
                iloc, jloc = i % (settings.nx // rs.num_proc[0]), j % (settings.ny // rs.num_proc[1])
                row.index = (i, j)

                for diag, offset in zip(cf, ij_offsets):
                    io, jo = (i + offset[0], j + offset[1])
                    col.index = (io, jo)
                    matrix.setValueStencil(row, col, diag[iloc, jloc])

        matrix.assemble()

        boundary_scale = {
            "east": npx.asarray(cf[1][-1, :]),
            "west": npx.asarray(cf[2][0, :]),
            "north": npx.asarray(cf[3][:, -1]),
            "south": npx.asarray(cf[4][:, 0]),
        }

        return matrix, boundary_scale


@veros_kernel
def prepare_solver_inputs(state, rhs, x0, boundary_val, boundary_fac):
    vs = state.variables
    settings = state.settings

    if boundary_val is None:
        boundary_val = x0

    x0 = utilities.enforce_boundaries(x0, settings.enable_cyclic_x)

    boundary_mask = ~npx.any(vs.boundary_mask, axis=2)
    rhs = npx.where(boundary_mask, rhs, boundary_val)  # set right hand side on boundaries

    # add dirichlet BC to rhs
    if not settings.enable_cyclic_x:
        if rst.proc_idx[0] == rs.num_proc[0] - 1:
            rhs = update_add(rhs, at[-3, 2:-2], -rhs[-2, 2:-2] * boundary_fac["east"])

        if rst.proc_idx[0] == 0:
            rhs = update_add(rhs, at[2, 2:-2], -rhs[1, 2:-2] * boundary_fac["west"])

    if rst.proc_idx[1] == rs.num_proc[1] - 1:
        rhs = update_add(rhs, at[2:-2, -3], -rhs[2:-2, -2] * boundary_fac["north"])

    if rst.proc_idx[1] == 0:
        rhs = update_add(rhs, at[2:-2, 2], -rhs[2:-2, 1] * boundary_fac["south"])

    return rhs, x0



import numpy as onp
import scipy.sparse
import scipy.sparse.linalg as spalg

from veros import logger, veros_kernel, veros_routine, distributed, runtime_state as rst
from veros.variables import allocate
from veros.core import utilities
from veros.core.operators import update, at, numpy as npx
from veros.core.streamfunction.solvers.base import LinearSolver


class SciPySolver(LinearSolver):
    @veros_routine(
        local_variables=("hvr", "hur", "dxu", "dxt", "dyu", "dyt", "cosu", "cost", "boundary_mask"),
        dist_safe=False,
    )
    def __init__(self, state):
        self._matrix = self._assemble_poisson_matrix(state)
        jacobi_precon = self._jacobi_preconditioner(state, self._matrix)
        self._matrix = jacobi_precon * self._matrix
        self._rhs_scale = jacobi_precon.diagonal()
        self._extra_args = {}

        logger.info("Computing ILU preconditioner...")
        ilu_preconditioner = spalg.spilu(self._matrix.tocsc(), drop_tol=1e-6, fill_factor=100)
        self._extra_args["M"] = spalg.LinearOperator(self._matrix.shape, ilu_preconditioner.solve)

    def _scipy_solver(self, state, rhs, x0, boundary_mask, boundary_val):
        settings = state.settings

        orig_shape = x0.shape
        orig_dtype = x0.dtype
        x0 = utilities.enforce_boundaries(x0, settings.enable_cyclic_x, local=True)

        rhs = npx.where(boundary_mask, rhs, boundary_val)  # set right hand side on boundaries

        rhs = onp.asarray(rhs.reshape(-1) * self._rhs_scale, dtype="float64")
        x0 = onp.asarray(x0.reshape(-1), dtype="float64")

        linear_solution, info = spalg.bicgstab(
            self._matrix,
            rhs,
            x0=x0,
            atol=1e-8,
            tol=0,
            maxiter=1000,
            **self._extra_args,
        )

        if info > 0:
            logger.warning("Streamfunction solver did not converge after {} iterations", info)

        return npx.asarray(linear_solution, dtype=orig_dtype).reshape(orig_shape)

    def solve(self, state, rhs, x0, boundary_val=None):
        """
        Main solver for streamfunction. Solves a 2D Poisson equation. Uses scipy.sparse.linalg
        linear solvers.

        Arguments:
            rhs: Right-hand side vector
            x0: Initial guess
            boundary_val: Array containing values to set on boundary elements. Defaults to `x0`.

        """
        rhs_global, x0_global, boundary_mask_global, boundary_val = gather_variables(state, rhs, x0, boundary_val)

        if rst.proc_rank == 0:
            linear_solution = self._scipy_solver(
                state, rhs_global, x0_global, boundary_mask=boundary_mask_global, boundary_val=boundary_val
            )
        else:
            linear_solution = npx.empty_like(rhs)

        return scatter_variables(state, linear_solution)

    @staticmethod
    def _jacobi_preconditioner(state, matrix):
        """
        Construct a simple Jacobi preconditioner
        """
        settings = state.settings

        eps = 1e-20
        precon = allocate(state.dimensions, ("xu", "yu"), fill=1, local=False)
        diag = npx.reshape(matrix.diagonal().copy(), (settings.nx + 4, settings.ny + 4))[2:-2, 2:-2]
        precon = update(precon, at[2:-2, 2:-2], npx.where(npx.abs(diag) > eps, 1.0 / (diag + eps), 1.0))
        precon = onp.array(precon)
        return scipy.sparse.dia_matrix((precon.reshape(-1), 0), shape=(precon.size, precon.size)).tocsr()

    @staticmethod
    def _assemble_poisson_matrix(state):
        """
        Construct a sparse matrix based on the stencil for the 2D Poisson equation.
        """
        vs = state.variables
        settings = state.settings

        boundary_mask = ~npx.any(vs.boundary_mask, axis=2)

        # assemble diagonals
        main_diag = allocate(state.dimensions, ("xu", "yu"), fill=1, local=False)
        east_diag, west_diag, north_diag, south_diag = (
            allocate(state.dimensions, ("xu", "yu"), local=False) for _ in range(4)
        )
        main_diag = update(
            main_diag,
            at[2:-2, 2:-2],
            -vs.hvr[3:-1, 2:-2]
            / vs.dxu[2:-2, npx.newaxis]
            / vs.dxt[3:-1, npx.newaxis]
            / vs.cosu[npx.newaxis, 2:-2] ** 2
            - vs.hvr[2:-2, 2:-2]
            / vs.dxu[2:-2, npx.newaxis]
            / vs.dxt[2:-2, npx.newaxis]
            / vs.cosu[npx.newaxis, 2:-2] ** 2
            - vs.hur[2:-2, 2:-2]
            / vs.dyu[npx.newaxis, 2:-2]
            / vs.dyt[npx.newaxis, 2:-2]
            * vs.cost[npx.newaxis, 2:-2]
            / vs.cosu[npx.newaxis, 2:-2]
            - vs.hur[2:-2, 3:-1]
            / vs.dyu[npx.newaxis, 2:-2]
            / vs.dyt[npx.newaxis, 3:-1]
            * vs.cost[npx.newaxis, 3:-1]
            / vs.cosu[npx.newaxis, 2:-2],
        )
        east_diag = update(
            east_diag,
            at[2:-2, 2:-2],
            vs.hvr[3:-1, 2:-2]
            / vs.dxu[2:-2, npx.newaxis]
            / vs.dxt[3:-1, npx.newaxis]
            / vs.cosu[npx.newaxis, 2:-2] ** 2,
        )
        west_diag = update(
            west_diag,
            at[2:-2, 2:-2],
            vs.hvr[2:-2, 2:-2]
            / vs.dxu[2:-2, npx.newaxis]
            / vs.dxt[2:-2, npx.newaxis]
            / vs.cosu[npx.newaxis, 2:-2] ** 2,
        )
        north_diag = update(
            north_diag,
            at[2:-2, 2:-2],
            vs.hur[2:-2, 3:-1]
            / vs.dyu[npx.newaxis, 2:-2]
            / vs.dyt[npx.newaxis, 3:-1]
            * vs.cost[npx.newaxis, 3:-1]
            / vs.cosu[npx.newaxis, 2:-2],
        )
        south_diag = update(
            south_diag,
            at[2:-2, 2:-2],
            vs.hur[2:-2, 2:-2]
            / vs.dyu[npx.newaxis, 2:-2]
            / vs.dyt[npx.newaxis, 2:-2]
            * vs.cost[npx.newaxis, 2:-2]
            / vs.cosu[npx.newaxis, 2:-2],
        )

        if settings.enable_cyclic_x:
            # couple edges of the domain
            wrap_diag_east, wrap_diag_west = (allocate(state.dimensions, ("xu", "yu"), local=False) for _ in range(2))
            wrap_diag_east = update(wrap_diag_east, at[2, 2:-2], west_diag[2, 2:-2] * boundary_mask[2, 2:-2])
            wrap_diag_west = update(wrap_diag_west, at[-3, 2:-2], east_diag[-3, 2:-2] * boundary_mask[-3, 2:-2])
            west_diag = update(west_diag, at[2, 2:-2], 0.0)
            east_diag = update(east_diag, at[-3, 2:-2], 0.0)

        main_diag = main_diag * boundary_mask
        main_diag = npx.where(main_diag == 0.0, 1.0, main_diag)

        # construct sparse matrix
        cf = tuple(
            diag.reshape(-1)
            for diag in (
                main_diag,
                boundary_mask * east_diag,
                boundary_mask * west_diag,
                boundary_mask * north_diag,
                boundary_mask * south_diag,
            )
        )
        offsets = (0, -main_diag.shape[1], main_diag.shape[1], -1, 1)

        if settings.enable_cyclic_x:
            offsets += (-main_diag.shape[1] * (settings.nx - 1), main_diag.shape[1] * (settings.nx - 1))
            cf += (wrap_diag_east.reshape(-1), wrap_diag_west.reshape(-1))

        cf = onp.asarray(cf, dtype="float64")
        return scipy.sparse.dia_matrix((cf, offsets), shape=(main_diag.size, main_diag.size)).T.tocsr()


@veros_kernel
def gather_variables(state, rhs, x0, boundary_val):
    vs = state.variables
    rhs_global = distributed.gather(rhs, state.dimensions, ("xt", "yt"))
    x0_global = distributed.gather(x0, state.dimensions, ("xt", "yt"))

    boundary_mask = ~npx.any(vs.boundary_mask, axis=2)
    boundary_mask_global = distributed.gather(boundary_mask, state.dimensions, ("xt", "yt"))

    if boundary_val is None:
        boundary_val = x0_global
    else:
        boundary_val = distributed.gather(boundary_val, state.dimensions, ("xt", "yt"))

    return rhs_global, x0_global, boundary_mask_global, boundary_val


@veros_kernel
def scatter_variables(state, linear_solution):
    return distributed.scatter(linear_solution, state.dimensions, ("xt", "yt"))
